import spiceypy
from spiceypy.utils.exceptions import SpiceyError

__all__ = ['Body']


class Body:
    """
    An individual body.

    Parameters
    ----------
    body : `int`, `str`
        Either the body ID code (integer) or the body name (string).
    """
    def __init__(self, body):
        if isinstance(body, int):
            self.id = body
        elif isinstance(body, str):
            self.name = body
        else:
            raise ValueError('body must be an int or str')

    def __repr__(self):
        return f'Body("{self.name}")'

    def __eq__(self, other):
        return isinstance(other, Body) and other.id == self.id

    @property
    def id(self):
        """Body ID code."""
        return self._id

    @id.setter
    def id(self, id):
        self._id = id
        try:
            self._name = spiceypy.bodc2n(id)
        except SpiceyError as e:
            raise ValueError(f'id "{id}" not known by SPICE') from e

    @property
    def name(self):
        """Body name."""
        return self._name

    @name.setter
    def name(self, name):
        self._name = name
        try:
            self._id = spiceypy.bodn2c(name)
        except SpiceyError as e:
            raise ValueError(f'Body name "{name}" not known by SPICE') from e



from urllib.request import urlopen

import astropy.units as u
import numpy as np
from astropy.time import Time
from bs4 import BeautifulSoup

from astrospice.net.reg import RemoteKernel, RemoteKernelsBase

__all__ = ['STEREORecon']


stereo_url = 'https://sohowww.nascom.nasa.gov/solarsoft/stereo/gen/data/spice'


class STEREORecon:
    type = 'recon'

    def get_remote_kernels(self):
        """
        Returns
        -------
        list[RemoteKernel]
        """
        url = f'{stereo_url}/depm/{self.spacecraft}/'
        page = urlopen(url)
        soup = BeautifulSoup(page, 'html.parser')

        kernel_urls = []
        for link in soup.find_all('a'):
            href = link.get('href')
            if href is not None and href.endswith('.bsp'):
                fname = href
                matches = self.matches(fname)
                if matches:
                    kernel_urls.append(
                        RemoteKernel(f'{url}{href}',
                                     *matches[1:]))

        for k1, k2 in zip(kernel_urls[:-1], kernel_urls[1:]):
            # Assume the end time of one kernel is the start time of the
            # subsequent kernel
            k1.end_time = k2.start_time

        # Mask the unknown end time of the most recent kernel
        kernel_urls[-1].end_time = Time(val=0, val2=np.nan, format='mjd')

        for k in kernel_urls:
            k.format = 'iso'
        return kernel_urls

    def matches(self, fname):
        """
        Check if the given filename matches the pattern of this kernel.

        Returns
        -------
        matches : bool
        start_time : astropy.time.Time
        end_time : astropy.time.Time
        version : int
        """
        # Example filename: ahead_2006_350_01.depm.bsp
        fname = fname.split('_')
        if (len(fname) != 4 or
                fname[0] != self.spacecraft):
            return False

        start_time = Time.strptime(fname[1] + fname[2], '%Y%j')
        end_time = None
        version = int(fname[3].split('.')[0])
        return True, start_time, end_time, version


class STEREOReconAhead(STEREORecon, RemoteKernelsBase):
    body = 'stereo-a'
    spacecraft = 'ahead'


class STEREOReconBehind(STEREORecon, RemoteKernelsBase):
    body = 'stereo-b'
    spacecraft = 'behind'


class STEREOPredAhead(RemoteKernelsBase):
    body = 'stereo-a'
    type = 'predict'

    fname = 'ahead_2017_061_5295day_predict.epm.bsp'

    def get_remote_kernels(self):
        start_time = Time.strptime('2017-061', '%Y-%j')
        end_time = start_time + 5295 * u.day
        return [RemoteKernel(f'{stereo_url}/epm/ahead/{self.fname}',
                             start_time, end_time, 1)]

    def matches(self, fname):
        return fname == self.fname



"""
`astrospice.registry`
---------------------
A registry of SPICE kernels for various missions.
"""
import abc
from collections import defaultdict
from dataclasses import dataclass

import astropy.time
import parfive
from astropy.table import Table, vstack
from astropy.time import Time

from astrospice.config import get_cache_dir
from astrospice.kernel import Kernel

__all__ = ['KernelRegistry', 'RemoteKernel', 'RemoteKernelsBase', 'registry']


class KernelRegistry:
    """
    A registry of remote SPICE kernels.

    Kernels are stored in the ``_kernels`` property, which is a dict with
    structure _kernels[body][type] -> RemoteKernelsBase
    """
    def __init__(self):
        self._kernels = defaultdict(dict)

    def __str__(self):
        return f'Known kernels: {self.bodies}'

    def __getitem__(self, idx):
        return self._kernels[idx]

    @property
    def bodies(self):
        """
        Bodies with available kernels.
        """
        return list(self._kernels.keys())

    def check_body(self, body):
        """
        Raise an error if ``body`` isn't in the registry.
        """
        if body not in self.bodies:
            raise ValueError(
                f'{body} not in list of registered bodies: {self.bodies}')

    def get_available_kernels(self, body):
        """
        Get a list of all the available kernels.

        Parameters
        ----------
        body : str
        type : str

        Returns
        -------
        astropy.table.Table
        """
        tables = []
        for type in self._kernels[body]:
            kernels = self._kernels[body][type].get_remote_kernels()
            urls = [k.url for k in kernels]
            stimes = Time([k.start_time for k in kernels])
            etimes = Time([k.end_time for k in kernels])
            versions = [k.version for k in kernels]
            tables.append(Table({'Mission': [body] * len(urls),
                                 'Type': [type] * len(urls),
                                 'Version': versions,
                                 'Start time': stimes,
                                 'End time': etimes}))

        tables = vstack(tables)
        tables['Start time'].format = 'iso'
        tables['End time'].format = 'iso'
        return tables

    def get_latest_kernel(self, body, type):
        """
        Returns
        -------
        astrospice.KernelBase
        """
        self.check_body(body)
        return self._kernels[body][type].get_latest_kernel()

    def get_kernels(self, body, type, *, version=None):
        """
        Download a set of kernels. Any kernels not present locally will be
        downloaded.

        Parameters
        ----------
        body : str
            Spacecraft body.
        type : str
            ``'recon'`` or ``'predict'`` to downloaded reconstructed or predicted
            kernels respectively.
        version : int, optional
            If given, get only kernels with this version.

        Returns
        -------
        list[pathlib.Path]
            List of local filepaths.
        """
        self.check_body(body)
        types = list(self._kernels[body].keys())
        if type not in types:
            raise ValueError(f'{type} is not one of the known kernel types '
                             f'for {body}: {types}')
        return self._kernels[body][type].get_kernels(version=version)


registry = KernelRegistry()


@dataclass
class RemoteKernel:
    """
    A single kernel available on a remote server.
    """
    url: str
    start_time: astropy.time.Time
    end_time: astropy.time.Time
    version: int

    def __str__(self):
        return '\n'.join([f'URL: {self.url}',
                          f'Start time: {self.start_time.iso}',
                          f'End time: {self.end_time.iso}',
                          f'Version: {self.version}'])

    def __lt__(self, other):
        return self.version < other.version

    @property
    def fname(self):
        """Kernel filename."""
        return self.url.split('/')[-1]

    def fetch(self):
        """
        Get the kernel. If not present locally, will be downloaded.

        This also implicitly furnishes the kernel with SPICE.

        Returns
        -------
        astrospice.KernelBase
        """
        local_path = get_cache_dir() / self.fname
        if not local_path.exists():
            dl = parfive.Downloader()
            dl.enqueue_file(self.url, get_cache_dir(), self.fname)
            dl.download()

        return Kernel(local_path)


class RemoteKernelsBase(abc.ABC):
    def __init_subclass__(cls):
        registry._kernels[cls.body][cls.type] = cls()
        assert cls.type in ['predict', 'recon']

    def get_latest_kernel(self):
        """
        Get the latest version of the kernel. If not present locally, will
        be downloaded.

        Returns
        -------
        astrospice.KernelBase
        """
        kernels = self.get_remote_kernels()
        k = sorted(kernels)[-1]
        return k.fetch()

    def get_kernels(self, *, version=None):
        """
        Get a set of kernels. Any kernels not present locally will be
        downloaded.

        Parameters
        ----------
        version : int, optional
            If given, get only this version of the kernel.
        trange : tuple[astropy.time.Time], optional
            If given, only get kernels to cover the given time range.

        Returns
        -------
        list[astrospice.KernelBase]
            List of kernels.

        Raises
        ------
        ValueError
            If there are no kernels available for the given type, version, and
            timerange.
        """
        kernels = self.get_remote_kernels()
        if version is not None:
            kernels = [k for k in kernels if k.version == version]

        if len(kernels) == 0:
            msg = f'No kernels available for {self.body}, type={self.type}'
            if version is not None:
                msg += f', version={version}'
            raise ValueError(msg)

        if self.type == 'predict':
            # Only get the most recent version
            kernels = [max(kernels)]
        dl = parfive.Downloader()
        for k in kernels:
            dl.enqueue_file(k.url, get_cache_dir(), k.fname)

        result = dl.download()
        return [Kernel(f) for f in result.data]

    @abc.abstractmethod
    def get_remote_kernels(self):
        """
        Get a list of all available remote kernels.

        Returns
        -------
        list[RemoteKernel]
        """



"""
Parker Solar Probe trajectory
=============================

In this example we compute and plot the trajectory of Parker Solar Probe across
it's nominal mission phase.
"""
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.time import Time, TimeDelta
from astropy.visualization import quantity_support
from sunpy.coordinates import HeliocentricInertial

import astrospice

# Enable support for plotting astropy Quantities.
quantity_support()

###############################################################################
# First we'll get the Parker Solar Probe SPICE kernel. To get the trajectory
# in the future choose the 'predict' option, which will give us a kernel for
# the predicted trajectory.
#
# We then inspect the kernel, to check it's time coverage.
kernels = astrospice.registry.get_kernels('psp', 'predict')
psp_kernel = kernels[0]
coverage = psp_kernel.coverage('SOLAR PROBE PLUS')
print(coverage.iso)

###############################################################################
# To generate some coordinates we do not need the kernel object as astrospice
# automatically registers the kernel with SPICE when ``get_kernels`` was called
# above.
dt = TimeDelta(0.5 * u.day)
times = Time(np.arange(coverage[0], Time('2022-01-01'), dt))
coords = astrospice.generate_coords('SOLAR PROBE PLUS', times)
print(coords[0:4])

###############################################################################
# Note that the coordinates are generated in the ICRS coordinate system. To
# change this to a more useful heliocentric coordinate system we can use
# sunpy's built in Heliocentric Intertial coordinate frame.
new_frame = HeliocentricInertial()
coords = coords.transform_to(new_frame)
print(coords[0:4])

###############################################################################
# Now we can plot the coordinates in this new coordinate system
fig = plt.figure()
ax = fig.add_subplot(projection='polar')
ax.scatter(coords.lon.to(u.rad), coords.distance.to(u.au), c=times.jd, s=2)
plt.show()



import logging
from pathlib import Path

import spiceypy
from astropy.time import Time

from astrospice.body import Body

__all__ = ['KernelBase', 'Kernel', 'SPKKernel', 'MetaKernel']


log = logging.getLogger(__name__)
# Mapping from filename extension to Kernel class
_REGISTRY = {}


class KernelBase:
    """
    Class for a single kernel.

    Notes
    -----
    When creating instances of this class, SPICE is automatically furnished
    with the kernel.
    """
    def __init__(self, fname):
        self._fname = fname
        spiceypy.furnsh(self._fname_str)

    def __init_subclass__(cls):
        _REGISTRY[cls._file_extension] = cls

    @property
    def fname(self):
        """Path to kernel file."""
        return Path(self._fname)

    @property
    def _fname_str(self):
        return str(self.fname)


def Kernel(fname):
    """
    Load a SPICE kernel.

    Parameters
    ----------
    fname : str, pathlib.Path
        Path to the kernel file.

    Returns
    -------
    kernel : KernelBase
        If the kernel type is known then this will be a sub-class of
        KernelBase.
    """
    extension = Path(fname).suffix
    if extension in _REGISTRY:
        return _REGISTRY[extension](fname)
    else:
        log.debug(f'Filename extension "{extension}" not in '
                  f'known extensions: {list(_REGISTRY.keys())}, '
                  'but furnishing with SPICE anyway.')
        return KernelBase(fname)


class SPKKernel(KernelBase):
    """
    A class for a single .spk kernel.

    References
    ----------
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/req/spk.html
    """
    _file_extension = '.bsp'

    def __init__(self, fname):
        super().__init__(fname)
        # Run bodies() to validate the spice kernel
        self.bodies

    def __str__(self):
        body_strs = []
        for b in self.bodies:
            body_strs.append(f'{b.name.capitalize()}')
        body_strs = ', '.join(body_strs)
        return f'SPK Kernel for {body_strs}'

    @property
    def bodies(self):
        """List of the bodies stored within the kernel."""
        ids = [int(i) for i in spiceypy.spkobj(self._fname_str)]
        return [Body(i) for i in ids]

    def coverage(self, body):
        """
        The coverage window for a specified `Body`.

        Parameters
        ----------
        body

        Returns
        -------
        astropy.time.Time
        """
        body = Body(body)
        coverage = [t for t in spiceypy.spkcov(self._fname_str, body.id)]
        return Time(coverage, format='et').utc


class MetaKernel(KernelBase):
    """
    A class for a single .tm kernel.

    References
    ----------
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/req/kernel.html#Additional%20Meta-kernel%20Specifications
    """
    _file_extension = '.tm'

    def __init__(self, fname):
        """
        Loading the metakernel will load all the kernels specified, if they exist.

        Parameters
        ----------
        fname : str, pathlib.Path
            Path to the metakernel file.
        """
        self._fname = fname
        if self.all_kernels_exist:
            self.load_kernels()
        else:
            print("Kernels are not yet loaded")

    @property
    def kernels(self):
        """
        The kernels specified by the metakernel

        Returns
        -------
        kernels : list of kernel Paths
        """
        kernels = []
        with open(self.fname, "r") as file:
            look_for_kernels = False
            for line in file:
                # split by whitespace
                line_split = line.split()

                # now find the ) bracket by itself, this is the end of kernels to load
                if len(line_split) >= 1 and line_split[0] == ')':
                    look_for_kernels = False
                    break

                if look_for_kernels and len(line_split) > 0:
                    # find the filename for the kernel.
                    # The slicing removes the $KERNEL and the final '
                    kernel_fname = line_split[0][9:-1]
                    # do not add empty lines
                    if kernel_fname != '':
                        # slice removes the first \
                        kernels.append(self.fname.parent / kernel_fname[1:])
                if len(line_split) > 1 and line_split[0] == 'KERNELS_TO_LOAD':
                    look_for_kernels = True
        return kernels

    def load_kernels(self):
        """
        Loads the kernels specified by the metakernel
        """
        for kernel in self.kernels:
            KernelBase(kernel)

    @property
    def all_kernels_exist(self):
        """
        Return `True` if all kernels in the metakernel exist.
        """
        for kernel in self.kernels:
            if not kernel.exists():
                return False
        return True




import astropy.units as u
import numpy as np
import spiceypy
from astropy.coordinates import SkyCoord
from astropy.time import Time

from astrospice.body import Body

__all__ = ['generate_coords']


def generate_coords(body, times):
    """
    Generate coordinates.

    Returns
    -------
    `~astropy.coordinates.SkyCoord`
    """
    body = Body(body)
    times = Time(times)
    times_et = np.atleast_1d(times.et)
    # Spice needs a funny set of times
    abcorr = str(None)
    frame = 'J2000'

    # Do the calculation
    pos_vel, lightTimes = spiceypy.spkezr(
        body.name, times_et, frame, abcorr,
        'SOLAR SYSTEM BARYCENTER')

    positions = np.array(pos_vel)[:, :3] * u.km
    velocities = np.array(pos_vel)[:, 3:] * u.km / u.s

    return SkyCoord(x=positions[:, 0],
                    y=positions[:, 1],
                    z=positions[:, 2],
                    obstime=times,
                    frame='icrs',
                    representation_type='cartesian')




import spiceypy
from astropy.coordinates.solar_system import solar_system_ephemeris
from astropy.utils.data import download_file

__all__ = ['set_solar_system_ephem', 'get_solar_system_ephem']

_known_jpl_ephem = ['de430', 'de432s', 'de440', 'de440s']
_jpl_ephem = solar_system_ephemeris.get()
if _jpl_ephem not in _known_jpl_ephem:
    _jpl_ephem = 'de440s'

_generic_files = [
    'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk',
]


def set_solar_system_ephem(name):
    """
    Set a JPL solar system ephemeris file. This will be downloaded and then
    cached for subsequent calls, and loaded into SPICE.

    Parameters
    ----------
    name : str
        Ephemeris name. Can be any ephemeris filename present at
        https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/
    """
    url = ('https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/'
           f'planets/{name}.bsp')
    fname = download_file(url, cache=True)
    spiceypy.furnsh(fname)

    global _jpl_ephem
    _jpl_ephem = name


def get_solar_system_ephem():
    """
    Get the currently set JPL solar system ephemeris file.

    Returns
    -------
    name : str
        Ephemeris name.
    """
    global _jpl_ephem
    return _jpl_ephem


def _setup_generic_files():
    for url in _generic_files:
        spiceypy.furnsh(download_file(url, cache=True))

    global _jpl_ephem
    set_solar_system_ephem(_jpl_ephem)



from urllib.request import urlopen

from astropy.time import Time
from bs4 import BeautifulSoup

from astrospice.net.reg import RemoteKernel, RemoteKernelsBase

__all__ = ['SolarOrbiterPredict']


class SolarOrbiterPredict(RemoteKernelsBase):
    body = 'solar orbiter'
    type = 'predict'

    def get_remote_kernels(self):
        """
        Returns
        -------
        list[RemoteKernel]
        """
        base_url = 'http://spiftp.esac.esa.int/data/SPICE/SOLAR-ORBITER/kernels/spk'
        page = urlopen(base_url)
        soup = BeautifulSoup(page, 'html.parser')

        kernel_urls = []
        for link in soup.find_all('a'):
            href = link.get('href')
            if href is not None and href.endswith('.bsp'):
                fname = href.split('/')[-1]
                matches = self.matches(fname)
                if matches:
                    kernel_urls.append(
                        RemoteKernel(f'{base_url}/{fname}',
                                     *matches[1:]))

        return kernel_urls

    @staticmethod
    def matches(fname):
        """
        Check if the given filename matches the pattern of this kernel.

        Returns
        -------
        matches : bool
        start_time : astropy.time.Time
        end_time : astropy.time.Time
        version : int
        """
        # Example filename: spp_nom_20180812_20250831_v038_RO5.bsp
        fname = fname.split('_')
        if (len(fname) != 8 or
                fname[0] != 'solo' or
                fname[1] != 'ANC' or
                fname[2] != 'soc-orbit'):
            return False

        times = fname[3].split('-')
        start_time = Time.strptime(times[0], '%Y%m%d')
        end_time = Time.strptime(times[1], '%Y%m%d')
        version = int(fname[4][1:])
        return True, start_time, end_time, version



from pathlib import Path

import pytest
from spiceypy.utils.exceptions import SpiceFILEREADFAILED, SpiceNOSUCHFILE

from astrospice import Kernel
from astrospice.kernel import MetaKernel

# mimic text structure of MetaKernel
METAKERNEL_CONTENT = "KERNELS_TO_LOAD   = (\n                           '$KERNELS/test_subfolder/test_kernel.bsp'\n                         )"


@pytest.fixture()
def example_mk(tmp_path):
    fname = tmp_path / "temp_mk.tm"
    fname.write_text(METAKERNEL_CONTENT)
    return fname


def create_example_kernel(tmp_path):
    extra_kernel = tmp_path / 'test_subfolder' / 'test_kernel.bsp'
    extra_kernel.parent.mkdir()
    extra_kernel.write_text(" ")
    return extra_kernel


def test_unkown_extension(tmp_path):
    fpath = tmp_path / 'test.xyz'

    with pytest.raises(SpiceNOSUCHFILE):
        Kernel(fpath)


def test_metakernel_kernels(example_mk):
    mk = MetaKernel(example_mk)
    kernels = mk.kernels
    assert isinstance(kernels, list)
    assert isinstance(kernels[0], Path)
    assert kernels[0].name == "test_kernel.bsp"
    assert kernels[0].parent.stem == 'test_subfolder'


def test_all_kernels_exist(example_mk, tmp_path):
    mk = MetaKernel(example_mk)
    assert not mk.all_kernels_exist
    create_example_kernel(tmp_path)
    assert mk.all_kernels_exist


def test_metakernels_load(example_mk, tmp_path):
    mk = MetaKernel(example_mk)
    create_example_kernel(tmp_path)
    # should get a SpiceFILEREADFAILED error
    with pytest.raises(SpiceFILEREADFAILED):
        mk.load_kernels()


def test_metakernels_init(example_mk, tmp_path):
    create_example_kernel(tmp_path)
    with pytest.raises(SpiceFILEREADFAILED):
        MetaKernel(example_mk)



from urllib.request import urlopen

from astropy.time import Time
from bs4 import BeautifulSoup

from astrospice.net.reg import RemoteKernel, RemoteKernelsBase

__all__ = ['CassiniRecon']


class CassiniRecon(RemoteKernelsBase):
    body = 'cassini'
    type = 'recon'

    def get_remote_kernels(self):
        """
        Returns
        -------
        list[RemoteKernel]
        """
        base_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/spk'
        page = urlopen(base_url)
        soup = BeautifulSoup(page, 'html.parser')

        kernel_urls = []
        for link in soup.find_all('a'):
            fname = link.get('href')
            if (fname is not None and fname.endswith('.bsp')):
                matches = self.matches(fname)
                if matches:
                    kernel_urls.append(RemoteKernel(
                        f'{base_url}/{fname}', *matches[1:]))

        return kernel_urls

    @staticmethod
    def matches(fname):
        """
        Check if the given filename matches the pattern of this kernel.

        Returns
        -------
        matches : bool
        start_time : astropy.time.Time
        end_time : astropy.time.Time
        version : int
        """
        # Example filename: 200128RU_SCPSE_09200_09215.bsp
        fname = fname.split('_')
        if (len(fname) != 4 or
                fname[0] != '200128RU' or
                fname[1] != 'SCPSE' or
                not fname[3].endswith('.bsp')):
            return False

        start_time = Time.strptime(fname[2], '%y%j')
        end_time = Time.strptime(fname[3].split('.')[0], '%y%j')
        version = 1
        return True, start_time, end_time, version



import pytest

from astrospice import Body


def test_errors():
    with pytest.raises(ValueError, match='body must be an int or str'):
        Body([])

    with pytest.raises(ValueError, match='id "45869" not known by SPICE'):
        Body(45869)

    msg = 'Body name "not a body" not known by SPICE'
    with pytest.raises(ValueError, match=msg):
        Body('not a body')



from pathlib import Path

import astropy.config.paths

__all__ = ['get_cache_dir']


def get_cache_dir():
    """
    Get path to the astrospice cache directory.

    Returns
    -------
    cachedir : pathlib.Path
        The absolute path to the cache directory.
    """
    return Path(astropy.config.paths.get_cache_dir(rootname='astrospice'))



import astropy
import astropy.units as u
import hypothesis.strategies as st
import pytest
from astropy.coordinates import get_body
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time
from hypothesis import given, settings
from packaging import version

import astrospice
from astrospice import generate_coords


@st.composite
def times(draw, min_time='1960-01-01', max_time='2024-01-01'):
    days = st.floats(min_value=0,
                     max_value=(Time(max_time) - Time(min_time)).to(u.day).value,
                     allow_nan=False, allow_infinity=False)
    return Time(min_time) + draw(days) * u.day


@pytest.mark.parametrize('ephem', ['de432s', 'de440s'])
@settings(max_examples=100, deadline=None)
@given(time=times())
def test_against_horizons(time, ephem):
    if (ephem == 'de440s' and
            (version.parse(astropy.__version__) < version.parse('4.3'))):
        pytest.skip('de440s ephemeris only available in astropy>=4.3')

    body = 'Earth'
    astrospice.set_solar_system_ephem(ephem)
    horizons_coord = get_body(body, time, ephemeris=ephem)
    astrospice_coord = generate_coords(body, time)
    assert_quantity_allclose(horizons_coord.separation_3d(astrospice_coord),
                             0*u.km, atol=50*u.m)


def test_different_ephem():
    # Check that setting a different ephemeris has an effect
    coords = []
    body = 'Earth'
    time = Time('2020-01-01')
    for ephem in ['de432s', 'de440s']:
        astrospice.set_solar_system_ephem(ephem)
        coords.append(generate_coords(body, time))

    # Check that generated coordinates are different
    assert coords[0].separation_3d(coords[1]) > 100 * u.m



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astrospice.time
from astrospice.net.generic import *
from astrospice.net.generic import _setup_generic_files
from ._version import __version__
from .body import *
from .config import *
from .coords import *
from .kernel import *
from .net import registry

_setup_generic_files()



from urllib.request import urlopen

from astropy.time import Time
from bs4 import BeautifulSoup

from astrospice.net.reg import RemoteKernel, RemoteKernelsBase

__all__ = ['PSPPredict', 'PSPRecon']


class PSPPredict(RemoteKernelsBase):
    body = 'psp'
    type = 'predict'

    def get_remote_kernels(self):
        """
        Returns
        -------
        list[RemoteKernel]
        """
        page = urlopen('https://spdf.gsfc.nasa.gov/pub/data/psp/ephemeris/spice/Long_Term_Predicted_Ephemeris/')
        soup = BeautifulSoup(page, 'html.parser')

        kernel_urls = []
        for link in soup.find_all('a'):
            href = link.get('href')
            if href is not None and href.startswith('spp'):
                fname = href.split('/')[-1]
                matches = self.matches(fname)
                if matches:
                    kernel_urls.append(
                        RemoteKernel(f'https://spdf.gsfc.nasa.gov/pub/data/psp/ephemeris/spice/Long_Term_Predicted_Ephemeris/{href}', *matches[1:]))

        return kernel_urls

    @staticmethod
    def matches(fname):
        """
        Check if the given filename matches the pattern of this kernel.

        Returns
        -------
        matches : bool
        start_time : astropy.time.Time
        end_time : astropy.time.Time
        version : int
        """
        # Example filename: spp_nom_20180812_20250831_v038_RO5.bsp
        fname = fname.split('_')
        if (len(fname) != 6 or
                fname[0] != 'spp' or
                fname[1] != 'nom'):
            return False

        start_time = Time.strptime(fname[2], '%Y%m%d')
        end_time = Time.strptime(fname[3], '%Y%m%d')
        version = int(fname[4][1:])
        return True, start_time, end_time, version


class PSPRecon(RemoteKernelsBase):
    body = 'psp'
    type = 'recon'

    def get_remote_kernels(self):
        """
        Returns
        -------
        list[RemoteKernel]
        """
        page = urlopen('https://sppgway.jhuapl.edu/recon_ephem')
        soup = BeautifulSoup(page, 'html.parser')

        kernel_urls = []
        for link in soup.find_all('a'):
            href = link.get('href')
            if (href is not None and
                    href.startswith('MOC/reconstructed_ephemeris') and
                    'archive' not in href):
                fname = href.split('/')[-1]
                matches = self.matches(fname)
                if matches:
                    kernel_urls.append(
                        RemoteKernel(f'https://sppgway.jhuapl.edu/{href}',
                                     *matches[1:]))

        return kernel_urls

    @staticmethod
    def matches(fname):
        """
        Check if the given filename matches the pattern of this kernel.

        Returns
        -------
        matches : bool
        start_time : astropy.time.Time
        end_time : astropy.time.Time
        version : int
        """
        # Example filename: spp_recon_20180812_20181008_v001.bsp
        fname = fname.split('_')
        if (len(fname) != 5 or
                fname[0] != 'spp' or
                fname[1] != 'recon'):
            return False

        start_time = Time.strptime(fname[2], '%Y%m%d')
        end_time = Time.strptime(fname[3], '%Y%m%d')
        version = int(fname[4].split('.')[0][1:])
        return True, start_time, end_time, version



import pytest
from astropy.table import Table

from astrospice import Body, SPKKernel, registry

# Ignore dubious year warnings for years a while in the future
pytestmark = pytest.mark.filterwarnings(r'ignore:ERFA function.*dubious year')


@pytest.mark.parametrize('body', registry.bodies)
def test_search(body):
    kernels = registry.get_available_kernels(body)
    assert isinstance(kernels, Table)


def test_get_kernels():
    k = registry.get_kernels('psp', 'predict', version=35)
    assert isinstance(k, list)
    assert len(k) == 1
    assert isinstance(k[0], SPKKernel)


def test_errors():
    with pytest.raises(ValueError, match='not a body not in list'):
        registry.get_kernels('not a body', 'recon')

    msg = 'blah is not one of the known kernel types for psp'
    with pytest.raises(ValueError, match=msg):
        registry.get_kernels('psp', 'blah')

    msg = 'No kernels available for psp, type=predict, version=1'
    with pytest.raises(ValueError, match=msg):
        registry.get_kernels('psp', 'predict', version=1)


def test_get_latest():
    kernel = registry.get_latest_kernel('psp', 'predict')
    assert isinstance(kernel, SPKKernel)
    assert kernel.bodies == [Body('SOLAR PROBE PLUS')]



from astropy.time.formats import TimeFromEpoch, erfa

__all__ = ['ETEpoch']


class ETEpoch(TimeFromEpoch):
    """
    Seconds from 12 noon, Jan 1st 2000. This is the epoch that SPICE uses
    internally.

    References
    ----------
    https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/req/time.html#The%20J2000%20Epoch
    """
    name = 'et'
    unit = 1.0 / erfa.DAYSEC  # Seconds
    epoch_val = '2000-01-01 12:00:00'
    epoch_val2 = None
    epoch_scale = 'tdb'
    epoch_format = 'iso'
